/*
    Copyright 2010 Luis Miguel Merino Bonilla

    This file is part of GNU Taste

    GNU Taste is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    GNU Taste is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with GNU Taste. If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef __KNN__
#define __KNN__

#include "database.h"
#include "hash.h"
#include "adjusted_cosine.hpp"

class KNN{
private:
	Storage *disk;
	Database *db;
	Hash< Hash<float> > sims;

public:
	KNN(Database *d) : db(d) {
		disk = new Storage();
		db->setNormalizing(true);
	}

	void normalizeItem(int i){
		for(node *n = db->getItem(i)->begin(); n != 0; n = n->next)
			n->value -= db->getMean(n->id);
	}

	//Similitud entre ítems
	float getSimilarity(int u, int v, Hash<float> &h){
		if(float *f=h.get(v))
			return *f;
		else	//No la tenemos en caché, calculamos
			return (sims[v][u] = h[v] = adjustedCosine(db->getItem(u), db->getItem(v), db));
	}

	//Calcula la tabla de similitud completa para un item
	Hash<float> & simTable(int item){
		Hash<float> & h = sims[item];
		feature_vector *this_item = db->getItem(item);
		//Para cada usuario que lo ha evaluado
		for(node *n = this_item->begin(); n != 0; n=n->next){
			//Para cada item evaluado por el usuario
			for(node *i = db->getUser(n->id)->begin(); i != 0; i=i->next){
				//Si no ha sido calculada antes
				if(!h.get(i->id)){
					//cout << "Calculando " << n->id << " " << i->id << endl;	
					//La calculamos
					h[i->id] = adjustedCosine(this_item, db->getItem(i->id), db);
				}
			}
		}

		return h;
	}

	//Predicción basada en ítems
	float predict(int user, int item){
		//cout << "Predict("<<user<<","<<item<<")" << endl;
		float num = 0, den = 0;
		feature_vector *f = db->getUser(user);
		Hash<float> & h = sims[item];

		int support = 0;

		for(node *n = f->begin(); n != 0; n=ne(n)){
			float sim = getSimilarity(item, n->id, h);
			if(sim > 0){
				num += sim * n->value;
				den += sim;
			}
			//cout << num << " " << den << endl;
			++support;
		}

		//cout << support << endl;

		if (den > 0)
			return num/den;
		else
			return 0;
	}
};

#endif
